//
// parser.cc
//
// Implementation of parser
//
// $Log$
//
#if RELEASE
static char RCSid[] = "$Id$";
#endif

#include "parser.h"

#define	WORD	1000
#define	DONE	1001


//*****************************************************************************
Parser::Parser()
{
	tokens = 0;
	result = 0;
	current = 0;
	dbf = 0;
	valid = 1;
}


//*****************************************************************************
// int Parser::checkSyntax(List *tokenList, Database *dbf)
//   As the name of the function implies, we will only perform a syntex check
//   on the list of tokens.
//
int
Parser::checkSyntax(List *tokenList)
{
	tokens = tokenList;
	tokens->Start_Get();
	lookahead = lexan();
	valid = 1;
	expr(0);
	return valid;
}

//*****************************************************************************
int
Parser::lexan()
{
	current = (WeightWord *) tokens->Get_Next();
	if (!current)
		return DONE;
	else if (mystrcasecmp(current->word, "&") == 0)
		return '&';
//	else if (mystrcasecmp(current->word, "and") == 0)
//		return '&';
	else if (mystrcasecmp(current->word, "|") == 0)
		return '|';
//	else if (mystrcasecmp(current->word, "or") == 0)
//		return '|';
	else if (mystrcasecmp(current->word, "(") == 0)
		return '(';
	else if (mystrcasecmp(current->word, ")") == 0)
		return ')';
	else
		return WORD;
}

//*****************************************************************************
//   Attempt to deal with expressions in the form
//		term | term | term ...
//
void
Parser::expr(int output)
{
	term(output);
	while (1)
	{
		if (match('|'))
		{
			term(output);
			if (output)
			{
				perform_or();
			}
		}
		else
			break;
	}
	if (lookahead == WORD)
	{
		valid = 0;
		error = 0;
		error << "expected 'AND' or 'OR' instead of '" << current->word.get();
		error << '\'';
	}
}

//*****************************************************************************
void
Parser::term(int output)
{
	factor(output);
	while (1)
	{
		if (match('&'))
		{
			factor(output);
			if (output)
			{
				perform_and();
			}
		}
		else
			break;
	}
}

//*****************************************************************************
void
Parser::factor(int output)
{
	if (match('('))
	{
		expr(output);
		if (match(')'))
		{
			return;
		}
		else
		{
			valid = 0;
		}
	}
	else if (lookahead == WORD)
	{
		if (output)
		{
			perform_push();
		}
		lookahead = lexan();
	}
	else
	{
		valid = 0;
	}
}

//*****************************************************************************
int
Parser::match(int t)
{
	if (lookahead == t)
	{
		lookahead = lexan();
		return 1;
	}
	else
		return 0;
}

//*****************************************************************************
// Perform a lookup of the current word and push the result onto the stack
//
void
Parser::perform_push()
{
	String		temp = current->word.get();
	String		data;
	char		*p;
	ResultList	*list = new ResultList;
	WordRecord	wr;
	DocMatch	*dm;

	stack.push(list);
	temp.lowercase();
	p = temp.get();
	if (temp.length() > MAX_WORD_LENGTH)
		p[MAX_WORD_LENGTH] = '\0';
	if (dbf->Get(p, data) == OK)
	{
		p = data.get();
		for (int i = 0; i < data.length() / sizeof(WordRecord); i++)
		{
			p = data.get() + i * sizeof(WordRecord);
			memcpy((char *) &wr, p, sizeof(WordRecord));

			//
			// *******  Compute the score for the document
			//
			dm = new DocMatch;
			dm->score = wr.weight * current->weight;
			dm->id = wr.id;
			dm->anchor = wr.anchor;
			list->add(dm);
		}
	}
}

//*****************************************************************************
// The top two entries in the stack need to be ORed together.
//
void
Parser::perform_and()
{
	ResultList		*l1 = (ResultList *) stack.pop();
	ResultList		*l2 = (ResultList *) stack.pop();
	ResultList		*result = new ResultList;
	int				i;
	DocMatch		*dm, *dm2, *dm3;
	List			*elements;

	stack.push(result);
	elements = l1->elements();
	for (i = 0; i < elements->Count(); i++)
	{
		dm = (DocMatch *) (*elements)[i];
		dm2 = l2->find(dm->id);
		if (dm2)
		{
			//
			// Duplicate document.  We just need to add the scored together.
			//
			dm3 = new DocMatch;
			dm3->score = dm->score + dm2->score;
			dm3->id = dm->id;
			dm3->anchor = dm->anchor;
			if (dm2->anchor < dm3->anchor)
				dm3->anchor = dm2->anchor;
			result->add(dm3);
		}
	}
	elements->Release();
	delete elements;
	delete l1;
	delete l2;
}

//*****************************************************************************
// The top two entries in the stack need to be ORed together.
//
void
Parser::perform_or()
{
	ResultList		*l1 = (ResultList *) stack.pop();
	ResultList		*result = (ResultList *) stack.peek();
	int				i;
	DocMatch		*dm, *dm2;
	List			*elements;

	elements = l1->elements();
	for (i = 0; i < elements->Count(); i++)
	{
		dm = (DocMatch *) (*elements)[i];
		dm2 = result->find(dm->id);
		if (dm2)
		{
			//
			// Duplicate document.  We just need to add the scores together
			//
			dm2->score += dm->score;
			if (dm->anchor < dm2->anchor)
				dm2->anchor = dm->anchor;
		}
		else
		{
			dm2 = new DocMatch;
			dm2->score = dm->score;
			dm2->id = dm->id;
			dm2->anchor = dm->anchor;
			result->add(dm2);
		}
	}
	elements->Release();
	delete elements;
	delete l1;
}

//*****************************************************************************
// void Parser::parse(List *tokenList, ResultList &resultMatches)
//
void
Parser::parse(List *tokenList, ResultList &resultMatches)
{
	tokens = tokenList;
	tokens->Start_Get();
	lookahead = lexan();
	expr(1);

	ResultList	*result = (ResultList *) stack.pop();
	List		*elements = result->elements();
	DocMatch	*dm;

	for (int i = 0; i < elements->Count(); i++)
	{
		dm = (DocMatch *) (*elements)[i];
		resultMatches.add(dm);
	}
	elements->Release();
	result->Release();
	delete elements;
	delete result;
}


